
package jcm.gui.plot;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;

import jcm.core.*;
import jcm.core.cur.curve;
import jcm.core.cur.curveset;
import jcm.core.cur.makederiv;
import jcm.core.cur.curvar;
import jcm.core.par.param;
import jcm.core.ob.*;
import jcm.core.cur.*;
import jcm.gui.doc.labman;
import jcm.gui.gen.colfont;
import jcm.gui.gen.lookandfeel;
import jcm.gui.nav.jcmMenu;
import jcm.gui.nav.paramValueChooser;

public class histoplot extends baseplot {
    
    public param year;
    public boolean highonleft=true;
    paramValueChooser yc; //=new paramValueChooser(year);
    java.util.List regs=new ArrayList();
    curveset qqx, qqy;
    curvar qqv;

    
    public histoplot(curveset qq) { this(new Object[]{qq}); }
    public histoplot(Object[] args) {
	for (Object o : args) {
	    if (o instanceof curvar) { qqv=(curvar)o;  qq=qqv; }
	    try { scalesetup.add(Float.parseFloat(o.toString())); } catch (NumberFormatException e) {}
	}
	qqy=qqv.linkqq.get(0); qqx=qqv.linkqq.get(1);
	double minsy=Math.max(qqy.sy, qqx.sy), maxey=Math.min(qqy.ey, qqx.ey);
	year =new param("year", 2002, minsy, maxey, "year");
	register.addlink(this, year);
	setup();
	sort();
	//ysv.setPreferredSize(new Dimension(60, 400));
    yc=new paramValueChooser(year);
	plot.add(yc, BorderLayout.NORTH);
    }

    public void hidechooser() { plot.remove(yc); }

    //******************************
    //override baseplot
    void makeplot() {   plot=new histo();  }
    
    public void doplot() { sort(); super.repaint(); }
//    public void setSize(Dimension d) { 	super.setSize(d); scale();     }
    
    void makescales() {
	for (Object o : qqx.mapwithouttotal().keySet()) regs.add(o);
	//for (Object o : qqq.get(1).map.keySet()) if (qqq.get(0).map.containsKey(o)) regs.add(o);
	float sum=0, max=0;
	for (Object o : regs) {sum+=qqx.getOrAddCurve(o).get((int)year.val); float f=calcratiotot(o);   if (f>max) max=f; }
	yscale=new param(param.Type.Yscale,  "Yscale", qqy.units+"&per&"+qqx.units,  0, 0, max);
	xscale=new param(param.Type.Xscale,  "Xscale", qqx.units,  0, 0, sum);
    }
    
    
    Comparator comp=new Comparator() {
	public int compare(Object a, Object b) {return (highonleft ? -1 : 1) * (calcratio(a) > calcratio(b) ? 1 : calcratio(a) < calcratio(b)  ? -1 : 0); }
    };
    
    public void fillMenu(jcmMenu popup) {
	popup.add(makederiv.makeDerivMenu(curve.Type.extra, qq)); popup.addSeparator();
	super.fillMenu(popup);
    }
    
    
//******************************
//RATIO
    float calcratio(Object o) {return calcratio(o, qqy); }
    float calcratio(Object o, curveset qqy) {
	float fr= qqx.getOrAddCurve(o).get((int)year.val); if (fr<=0) return 0;
	return qqy.getOrAddCurve(o).get((int)year.val) / fr;
    }
    
    float calcratiotot(Object o) {
	float f=0;  for (curveset qqy : qqv.linkqq) if(qqy!=qqx) f+=calcratio(o, qqy); return f;
    }
    
    void sort() {
	Collections.sort(regs, comp);
	//for (Object o : regs) System.err.println(o.toString()+ " "+calcratio(o,1)); //" "+calcratio(o,2)+" "+calctotratio(o));
    }
    
    
    
//**************************
//PLOT HISTOGRAM
    class histo extends JPanel implements  MouseMotionListener {
	public  histo() {
	    addMouseMotionListener(this);
	}
	
	public void paintComponent(Graphics g) {
	    super.paintComponent(g);
	    lookandfeel.setAntiAlias(g);
        g.setFont(colfont.normalfont);
        int w=getSize().width, h=getSize().height;
	    
	    float xx=-xscale.min()*w/xscale.range(), x, yy, y;
	    for (Object o : regs) {
		if (o instanceof infob) g.setColor(((infob)o).getColor()) ;
		x=(qqx.getOrAddCurve(o).get((int)year.val)*w/xscale.range());
		yy=h;
		
		for (curveset qqy : qqv.linkqq) if(qqy!=qqx) {
		    y=(int)(calcratio(o, qqy)*h / yscale.max());
		    if (y<0) {
			g.setColor(Color.white);
			g.fillRect((int)xx+1, (int)yy+1, (int)x-2,  (int)-y-1);
		    } else g.fillRect((int)xx, (int)(yy-y)-1, (int)x,  (int)y-1);
		    g.setColor(variant(g.getColor()));
		    yy-=y;
		}
        
        if (labels==true)  {
            if (o instanceof infob) g.setColor(((infob)o).getColor()) ;
            String s=labman.getShort((o instanceof infob) ? ((infob)o).name : o.toString());
            g.setFont(colfont.smallfont);
            int sw=colfont.sw(g, s), sh=colfont.sh(g), sd=colfont.sd(g);
            int sx=(int)(xx+x/2-sw/2), sy=(int)(yy-sh-4);
            //if (sw>x) { g.setFont(colfont.smallfont); sw=colfont.sw(g, s); sh=colfont.sh(g); }
           // g.drawString(s, (int)(xx+x/2-sw/2), (int)(yy-sh-4));
            Graphics2D g2=(Graphics2D) g.create();
            g2.translate(xx+x/2, yy-4);
            g2.rotate(-Math.PI/2.0);
            g2.drawString(s, 0, sh/2-sd);
            g2.dispose();
        }
		
		xx+=x;
	     /*
	    y=(int)(calcratio(o, 1)*h/max);
	    g.fillRect((int)xx, (int)(yy-y), (int)x,  (int)y);
	    yy-=y;
	    y=(int)(calcratio(o, 2)*h/max);
	    if (y<0) g.setColor(Color.white);
	    g.fillPolygon(new int[]{(int)xx, (int)(xx+x), (int)(xx+x/2) }, new int[] {(int) yy, (int)yy, (int)(yy-y)}, 3 );
	      */
		
	    }
	    ready=true;
        
	} //paint
	
	
	
	//**************************
	//MOUSE
	public void mouseMoved(MouseEvent e) {
	    float xx=0, x;
	    findreg: for (Object o : regs) {
		x=(qqx.getOrAddCurve(o).get((int)year.val)*getSize().width/xscale.max());
		if (xx+x > e.getX() && o instanceof infob) {
		    String s= "<html>"+((infob)o).hashcolor() +((infob)o).getName();
		    for (curveset qqy : qqv.linkqq) if(qqy!=qqx) s+=" "+calcratio(o, qqy);
		    setToolTipText(s);
		    return;
		} else xx+=x;
	    }
	}
	
	public void mouseDragged(MouseEvent e) {	}
	
    } //histo class
    
}

//SortedMap<Object, Float> ratio=new TreeMap();
    /*
    class ratio implements Comparable<ratio> {
	Object o; Float f;
	ratio(Object o, Float f) {this.o=o; this.f=f;}
	public int compareTo(ratio r) { return (int)(f-r.f); }
    }
     */

